"""Helper for OpenHAB items."""

from __future__ import annotations

import contextlib
import datetime

import HABApp.core
import HABApp.openhab.events

NO_VALUE = object()
_MOCKED_ITEM_NAMES = []
StateTypes = str | float | datetime.datetime


def add_mock_item(item_type: type[HABApp.openhab.items.OpenhabItem], name: str, initial_value: str | float | None = None) -> None:
    """Add a mock item.

    Args:
        item_type: Type of the mock item
        name: Name of the mock item
        initial_value: initial value
    """
    if HABApp.core.Items.item_exists(name):
        HABApp.core.Items.pop_item(name)
    item = item_type(name, initial_value)
    HABApp.core.Items.add_item(item)
    _MOCKED_ITEM_NAMES.append(name)


def remove_mocked_item_by_name(name: str) -> None:
    """Remove a mocked item by item name.

    Args:
        name: name of mocked item
    """
    HABApp.core.Items.pop_item(name)
    _MOCKED_ITEM_NAMES.remove(name)


def remove_all_mocked_items() -> None:
    """Remove all mocked items."""
    for name in _MOCKED_ITEM_NAMES:
        HABApp.core.Items.pop_item(name)
    _MOCKED_ITEM_NAMES.clear()


def set_state(item_name: str, value: StateTypes | None) -> None:
    """Helper to set state of item.

    Args:
        item_name: name of item
        value: state which should be set
    """
    item = HABApp.openhab.items.OpenhabItem.get_item(item_name)
    if isinstance(item, HABApp.openhab.items.DimmerItem) and value in {"ON", "OFF"}:
        value = 100 if value == "ON" else 0

    with contextlib.suppress(AssertionError):
        item.set_value(value)


def send_command(item_name: str, new_value: StateTypes, old_value: StateTypes = NO_VALUE) -> None:
    """Replacement of send_command for unit-tests.

    Args:
        item_name: Name of item
        new_value: new value
        old_value: previous value
    """
    old_value = HABApp.openhab.items.OpenhabItem.get_item(item_name).value if old_value is NO_VALUE else old_value

    set_state(item_name, new_value)
    if old_value is not NO_VALUE and old_value != new_value:
        HABApp.core.EventBus.post_event(item_name, HABApp.openhab.events.ItemStateChangedEvent(item_name, new_value, old_value))
    HABApp.core.EventBus.post_event(item_name, HABApp.openhab.events.ItemStateUpdatedEvent(item_name, new_value))


def item_command_event(item_name: str, value: StateTypes) -> None:
    """Post a command event to the event bus.

    Args:
        item_name: name of item
        value: value of the event
    """
    with contextlib.suppress(HABApp.core.errors.InvalidItemValue):
        set_state(item_name, value)
    HABApp.core.EventBus.post_event(item_name, HABApp.openhab.events.ItemCommandEvent(item_name, value))


def item_state_event(item_name: str, value: StateTypes) -> None:
    """Post a state event to the event bus.

    Args:
        item_name: name of item
        value: value of the event
    """
    set_state(item_name, value)
    HABApp.core.EventBus.post_event(item_name, HABApp.openhab.events.ItemStateUpdatedEvent(item_name, value))


def item_state_change_event(item_name: str, value: StateTypes, old_value: StateTypes = None) -> None:
    """Post a state change event to the event bus.

    Args:
        item_name: name of item
        value: value of the event
        old_value: previous value
    """
    prev_value = old_value or HABApp.openhab.items.OpenhabItem.get_item(item_name).value
    set_state(item_name, value)
    HABApp.core.EventBus.post_event(item_name, HABApp.openhab.events.ItemStateChangedEvent(item_name, value, prev_value))


def assert_value(item_name: str, value: StateTypes | None, message: str | None = None) -> None:
    """Helper to assert if item has correct state.

    Args:
        item_name: name of item
        value: expected state
        message: message to display if assertion failed

    Raises:
        AssertionError: if value is wrong
    """
    if (current_state := HABApp.openhab.items.OpenhabItem.get_item(item_name).value) != value:
        msg = f"Wrong state of item '{item_name}'. Expected: {value} | Current: {current_state}"
        if message:
            msg += f"message = {message}"
        raise AssertionError(msg)



"""Config models for ventilation rules."""

import datetime

import HABApp
import pydantic

import habapp_rules.core.pydantic_base


class StateConfig(pydantic.BaseModel):
    """Basic state config."""

    level: int
    display_text: str


class StateConfigWithTimeout(StateConfig):
    """State config with timeout."""

    timeout: int


class StateConfigLongAbsence(StateConfig):
    """State config for long absence state."""

    duration: int = 3600
    start_time: datetime.time = datetime.time(6)


class _VentilationItemsBase(habapp_rules.core.pydantic_base.ItemBase):
    """Base class for ventilation items."""

    manual: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="Item to disable all automatic functions")
    hand_request: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="Item to enter the hand state")
    external_request: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="Item to enter the external state")
    presence_state: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="Item of presence state to detect long absence")
    feedback_on: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="Item which shows that ventilation is on")
    feedback_power: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="Item which shows that ventilation is in power mode")
    display_text: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="Item which can be used to set the display text")
    state: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="Item for storing the current state")


class VentilationItems(_VentilationItemsBase):
    """Items for ventilation."""

    ventilation_level: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="Item to set the ventilation level")


class VentilationTwoStageItems(_VentilationItemsBase):
    """Items for ventilation."""

    ventilation_output_on: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="Item to switch on the ventilation")
    ventilation_output_power: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="Item to switch on the power mode")
    current: HABApp.openhab.items.NumberItem | None = pydantic.Field(None, description="Item to measure the current of the ventilation")
    feedback_ventilation_level: HABApp.openhab.items.NumberItem | None = pydantic.Field(None, description="Item feedback current ventilation level")


class VentilationParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for ventilation."""

    state_normal: StateConfig = pydantic.Field(StateConfig(level=1, display_text="Normal"))
    state_hand: StateConfigWithTimeout = pydantic.Field(StateConfigWithTimeout(level=2, display_text="Hand", timeout=3600))
    state_external: StateConfig = pydantic.Field(StateConfig(level=2, display_text="External"))
    state_humidity: StateConfig = pydantic.Field(StateConfig(level=2, display_text="Humidity"))
    state_long_absence: StateConfigLongAbsence = pydantic.Field(StateConfigLongAbsence(level=2, display_text="LongAbsence"))


class VentilationTwoStageParameter(VentilationParameter):
    """Parameter for ventilation."""

    state_after_run: StateConfig = pydantic.Field(StateConfig(level=2, display_text="After run"))
    after_run_timeout: int = pydantic.Field(390, description="")
    current_threshold_power: float = pydantic.Field(0.105, description="")


class VentilationConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for ventilation."""

    items: VentilationItems = pydantic.Field(..., description="Items for ventilation")
    parameter: VentilationParameter = pydantic.Field(VentilationParameter(), description="Parameter for ventilation")


class VentilationTwoStageConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for ventilation."""

    items: VentilationTwoStageItems = pydantic.Field(..., description="Items for ventilation")
    parameter: VentilationTwoStageParameter = pydantic.Field(VentilationTwoStageParameter(), description="Parameter for ventilation")



"""Config models for shading rules."""

from __future__ import annotations

import copy

import HABApp.openhab.items  # noqa: TC002
import pydantic
import typing_extensions

import habapp_rules.core.pydantic_base


class ShadingPosition(pydantic.BaseModel):
    """Position of shading object."""

    position: float | bool | None = pydantic.Field(..., description="target position")
    slat: float | None = pydantic.Field(None, description="target slat position")

    def __init__(self, position: float | bool | None, slat: float | None = None) -> None:
        """Initialize shading position with position and slat.

        Args:
            position: target position value
            slat: slat value
        """
        super().__init__(position=position, slat=slat)


class ShadingItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for shading rules."""

    shading_position: HABApp.openhab.items.RollershutterItem | HABApp.openhab.items.DimmerItem = pydantic.Field(..., description="item for setting the shading position")
    slat: HABApp.openhab.items.DimmerItem | None = pydantic.Field(None, description="item for setting the slat value")
    manual: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item to switch to manual mode and disable the automatic functions")
    shading_position_control: list[HABApp.openhab.items.RollershutterItem | HABApp.openhab.items.DimmerItem] = pydantic.Field([], description="control items to improve manual detection")
    shading_position_group: list[HABApp.openhab.items.RollershutterItem | HABApp.openhab.items.DimmerItem] = pydantic.Field([], description="")
    wind_alarm: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="item which is ON when wind alarm is active")
    sun_protection: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="item which is ON when sun protection is needed")
    sun_protection_slat: HABApp.openhab.items.DimmerItem | None = pydantic.Field(None, description="value for the slat when sun protection is active")
    sleeping_state: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="item of the sleeping state set via habapp_rules.system.sleep.Sleep")
    night: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="item which is ON at night or darkness")
    door: HABApp.openhab.items.ContactItem | None = pydantic.Field(None, description="item for setting position when door is opened")
    summer: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="item which is ON during summer")
    hand_manual_is_active_feedback: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="feedback item which is ON when hand or manual is active")
    state: HABApp.openhab.items.StringItem = pydantic.Field(..., description="item to store the current state of the state machine")


class ShadingParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for shading rules."""

    pos_auto_open: ShadingPosition = pydantic.Field(ShadingPosition(0, 0), description="position for auto open")
    pos_wind_alarm: ShadingPosition | None = pydantic.Field(ShadingPosition(0, 0), description="position for wind alarm")
    pos_sleeping_night: ShadingPosition | None = pydantic.Field(ShadingPosition(100, 100), description="position for sleeping at night")
    pos_sleeping_day: ShadingPosition | None = pydantic.Field(None, description="position for sleeping at day")
    pos_sun_protection: ShadingPosition | None = pydantic.Field(ShadingPosition(100, None), description="position for sun protection")
    pos_night_close_summer: ShadingPosition | None = pydantic.Field(None, description="position for night close during summer")
    pos_night_close_winter: ShadingPosition | None = pydantic.Field(ShadingPosition(100, 100), description="position for night close during winter")
    pos_door_open: ShadingPosition | None = pydantic.Field(ShadingPosition(0, 0), description="position if door is opened")
    manual_timeout: int = pydantic.Field(24 * 3600, description="fallback timeout for manual state", gt=0)
    door_post_time: int = pydantic.Field(5 * 60, description="extended time after door is closed", gt=0)
    value_tolerance: int = pydantic.Field(0, description="value tolerance for shading position which is allowed without manual detection", ge=0)

    @pydantic.model_validator(mode="after")
    def validate_model(self) -> typing_extensions.Self:
        """Validate model.

        Returns:
            validated model
        """
        if self.pos_sleeping_night and not self.pos_sleeping_day:
            self.pos_sleeping_day = copy.deepcopy(self.pos_sleeping_night)
        return self


class ShadingConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for shading objects."""

    items: ShadingItems = pydantic.Field(..., description="items for shading")
    parameter: ShadingParameter = pydantic.Field(ShadingParameter(), description="parameter for shading")

    @pydantic.model_validator(mode="after")
    def validate_model(self) -> typing_extensions.Self:
        """Validate model.

        Returns:
            validated model

        Raises:
            AssertionError: if 'parameter.pos_night_close_summer' is set but 'items.summer' is missing
        """
        if self.parameter.pos_night_close_summer is not None and self.items.summer is None:
            msg = "Night close position is set for summer, but item for summer / winter is missing!"
            raise AssertionError(msg)
        return self


class ResetAllManualHandItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for reset all manual hand items."""

    reset_manual_hand: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item for resetting manual and hand state to automatic state")


class ResetAllManualHandParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for reset all manual hand parameter."""

    shading_objects: list[object] | None = pydantic.Field(None, description="list of shading objects to reset")


class ResetAllManualHandConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for reset all manual hand config."""

    items: ResetAllManualHandItems = pydantic.Field(..., description="items for reset all manual hand config")
    parameter: ResetAllManualHandParameter = pydantic.Field(ResetAllManualHandParameter(), description="parameter for reset all manual hand config")


class SlatValueItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for slat values for sun protection."""

    sun_elevation: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="item for sun elevation")
    slat_value: HABApp.openhab.items.NumberItem | HABApp.openhab.items.DimmerItem = pydantic.Field(..., description="item for slat value, which should be set")
    summer: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="item for summer")


class ElevationSlatMapping(pydantic.BaseModel):
    """Mapping from elevation to slat value."""

    elevation: int
    slat_value: int

    def __init__(self, elevation: int, slat_value: int) -> None:
        """Initialize the elevation slat mapping.

        Args:
            elevation: elevation value
            slat_value: mapped slat value
        """
        super().__init__(elevation=elevation, slat_value=slat_value)


class SlatValueParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for slat values for sun protection."""

    elevation_slat_characteristic: list[ElevationSlatMapping] = pydantic.Field(
        [ElevationSlatMapping(0, 100), ElevationSlatMapping(4, 100), ElevationSlatMapping(8, 90), ElevationSlatMapping(18, 80), ElevationSlatMapping(26, 70), ElevationSlatMapping(34, 60), ElevationSlatMapping(41, 50)],
        description="list of tuple-mappings from elevation to slat value",
    )
    elevation_slat_characteristic_summer: list[ElevationSlatMapping] = pydantic.Field(
        [ElevationSlatMapping(0, 100), ElevationSlatMapping(4, 100), ElevationSlatMapping(8, 90), ElevationSlatMapping(18, 80)], description="list of tuple-mappings from elevation to slat value, which is used if summer is active"
    )

    @pydantic.field_validator("elevation_slat_characteristic", "elevation_slat_characteristic_summer")
    @classmethod
    def sort_mapping(cls, values: list[ElevationSlatMapping]) -> list[ElevationSlatMapping]:
        """Sort the elevation slat mappings.

        Args:
            values: input values

        Returns:
            sorted values

        Raises:
            AssertionError: if elevation values are not unique
        """
        values.sort(key=lambda x: x.elevation)

        if len(values) != len({value.elevation for value in values}):
            msg = "Elevation values must be unique!"
            raise AssertionError(msg)

        return values


class SlatValueConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for slat values for sun protection."""

    items: SlatValueItems = pydantic.Field(..., description="items for slat values for sun protection")
    parameter: SlatValueParameter = pydantic.Field(SlatValueParameter(), description="parameter for slat values for sun protection")


class ReferenceRunItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for reference run."""

    trigger_run: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item for triggering the reference run")
    last_run: HABApp.openhab.items.DatetimeItem = pydantic.Field(..., description="item for date/time of the last run")
    presence_state: HABApp.openhab.items.StringItem = pydantic.Field(..., description="item for presence state")


class ReferenceRunConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for reference run."""

    items: ReferenceRunItems = pydantic.Field(..., description="items for reference run")
    parameter: None = None



"""Test light HCL rules."""

import collections
import datetime
import pathlib
import sys
import unittest.mock

import HABApp

import habapp_rules.actors.config.light_hcl
import habapp_rules.actors.light_hcl
import habapp_rules.core.exceptions
import habapp_rules.system
import tests.helper.graph_machines
import tests.helper.oh_item
import tests.helper.test_case_base


class TestHclElevation(tests.helper.test_case_base.TestCaseBaseStateMachine):
    """Tests for elevation-based HCL."""

    def setUp(self) -> None:
        """Setup tests."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Elevation", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Color_min", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Manual_min", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_Unittest_Color_min_state", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Color_max", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Manual_max", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Sleep_state", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Focus_max", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch_on_max", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_State_max", None)

        self._config_min = habapp_rules.actors.config.light_hcl.HclElevationConfig(
            items=habapp_rules.actors.config.light_hcl.HclElevationItems(color="Unittest_Color_min", manual="Unittest_Manual_min", elevation="Unittest_Elevation", state="H_Unittest_Color_min_state"),
            parameter=habapp_rules.actors.config.light_hcl.HclElevationParameter(color_map=[(-10, 3000), (-2, 3800), (0, 4200.0), (10, 5000)]),
        )

        self._config_max = habapp_rules.actors.config.light_hcl.HclElevationConfig(
            items=habapp_rules.actors.config.light_hcl.HclElevationItems(
                color="Unittest_Color_max",
                manual="Unittest_Manual_max",
                elevation="Unittest_Elevation",
                state="H_State_max",
                sleep_state="Unittest_Sleep_state",
                focus="Unittest_Focus_max",
                switch_on="Unittest_Switch_on_max",
            ),
            parameter=habapp_rules.actors.config.light_hcl.HclElevationParameter(color_map=[(-10, 3000), (-2, 3800), (0, 4200.0), (10, 5000)], hand_timeout=30 * 60, sleep_color=3000, post_sleep_timeout=500, focus_color=7000),
        )

        self._hcl_elevation_min = habapp_rules.actors.light_hcl.HclElevation(self._config_min)
        self._hcl_elevation_max = habapp_rules.actors.light_hcl.HclElevation(self._config_max)

    @unittest.skipIf(sys.platform != "win32", "Should only run on windows when graphviz is installed")
    def test_create_graph(self) -> None:  # pragma: no cover
        """Create state machine graph for documentation."""
        picture_dir = pathlib.Path(__file__).parent / "_state_charts" / "Light_HCL"
        if not picture_dir.is_dir():
            picture_dir.mkdir(parents=True)

        graph = tests.helper.graph_machines.HierarchicalGraphMachineTimer(
            model=tests.helper.graph_machines.FakeModel(), states=self._hcl_elevation_min.states, transitions=self._hcl_elevation_min.trans, initial=self._hcl_elevation_min.state, show_conditions=True
        )

        graph.get_graph().draw(picture_dir / "HCL_Base.png", format="png", prog="dot")

    def test_set_timeouts(self) -> None:
        """Test _set_timeouts."""
        # min
        self.assertEqual(18000, self._hcl_elevation_min.state_machine.get_state("Hand").timeout)
        self.assertEqual(1, self._hcl_elevation_min.state_machine.get_state("Auto_Sleep_Post").timeout)

        # max
        self.assertEqual(1800, self._hcl_elevation_max.state_machine.get_state("Hand").timeout)
        self.assertEqual(500, self._hcl_elevation_max.state_machine.get_state("Auto_Sleep_Post").timeout)

    def test_get_initial_state(self) -> None:
        """Test _get_initial_state."""
        TestCase = collections.namedtuple("TestCase", "manual, focus, sleep_state, result_min, result_max")

        test_cases = [
            TestCase("OFF", "OFF", habapp_rules.system.SleepState.AWAKE, "Auto_HCL", "Auto_HCL"),
            TestCase("OFF", "OFF", habapp_rules.system.SleepState.PRE_SLEEPING, "Auto_HCL", "Auto_Sleep"),
            TestCase("OFF", "OFF", habapp_rules.system.SleepState.SLEEPING, "Auto_HCL", "Auto_Sleep"),
            TestCase("OFF", "OFF", habapp_rules.system.SleepState.POST_SLEEPING, "Auto_HCL", "Auto_HCL"),
            TestCase("OFF", "ON", habapp_rules.system.SleepState.AWAKE, "Auto_HCL", "Auto_Focus"),
            TestCase("OFF", "ON", habapp_rules.system.SleepState.PRE_SLEEPING, "Auto_HCL", "Auto_Sleep"),
            TestCase("OFF", "ON", habapp_rules.system.SleepState.SLEEPING, "Auto_HCL", "Auto_Sleep"),
            TestCase("OFF", "ON", habapp_rules.system.SleepState.POST_SLEEPING, "Auto_HCL", "Auto_Focus"),
            TestCase("ON", "OFF", habapp_rules.system.SleepState.AWAKE, "Manual", "Manual"),
            TestCase("ON", "OFF", habapp_rules.system.SleepState.PRE_SLEEPING, "Manual", "Manual"),
            TestCase("ON", "OFF", habapp_rules.system.SleepState.SLEEPING, "Manual", "Manual"),
            TestCase("ON", "OFF", habapp_rules.system.SleepState.POST_SLEEPING, "Manual", "Manual"),
            TestCase("ON", "ON", habapp_rules.system.SleepState.AWAKE, "Manual", "Manual"),
            TestCase("ON", "ON", habapp_rules.system.SleepState.PRE_SLEEPING, "Manual", "Manual"),
            TestCase("ON", "ON", habapp_rules.system.SleepState.SLEEPING, "Manual", "Manual"),
            TestCase("ON", "ON", habapp_rules.system.SleepState.POST_SLEEPING, "Manual", "Manual"),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.set_state("Unittest_Manual_min", test_case.manual)
                tests.helper.oh_item.set_state("Unittest_Manual_max", test_case.manual)
                tests.helper.oh_item.set_state("Unittest_Focus_max", test_case.focus)
                tests.helper.oh_item.set_state("Unittest_Sleep_state", test_case.sleep_state.value)

                self.assertEqual(test_case.result_min, self._hcl_elevation_min._get_initial_state())
                self.assertEqual(test_case.result_max, self._hcl_elevation_max._get_initial_state())

    def test_get_hcl_color(self) -> None:
        """Test _get_hcl_color."""
        TestCase = collections.namedtuple("TestCase", "input, output")

        test_cases = [
            TestCase(-20, 3000),
            TestCase(-10.5, 3000),
            TestCase(-10, 3000),
            TestCase(-6, 3400),
            TestCase(-2, 3800),
            TestCase(-1, 4000),
            TestCase(0, 4200),
            TestCase(5, 4600),
            TestCase(10, 5000),
            TestCase(12, 5000),
            TestCase(0.5, 4240),
            TestCase(0.2556, 4220),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self._hcl_elevation_min._config.items.elevation.value = test_case.input
                self.assertEqual(test_case.output, self._hcl_elevation_min._get_hcl_color())

    def test_end_to_end(self) -> None:
        """Test end to end behavior."""
        tests.helper.oh_item.assert_value("Unittest_Color_min", None)
        tests.helper.oh_item.item_state_change_event("Unittest_Elevation", 0)
        tests.helper.oh_item.assert_value("Unittest_Color_min", 4200)

    def test_manual(self) -> None:
        """Test manual."""
        self.assertEqual("Auto_HCL", self._hcl_elevation_min.state)
        self.assertEqual("Auto_HCL", self._hcl_elevation_max.state)

        tests.helper.oh_item.item_state_change_event("Unittest_Manual_min", "ON")
        tests.helper.oh_item.item_state_change_event("Unittest_Manual_max", "ON")
        self.assertEqual("Manual", self._hcl_elevation_min.state)
        self.assertEqual("Manual", self._hcl_elevation_max.state)

        tests.helper.oh_item.item_state_change_event("Unittest_Manual_min", "OFF")
        tests.helper.oh_item.item_state_change_event("Unittest_Manual_max", "OFF")
        self.assertEqual("Auto_HCL", self._hcl_elevation_min.state)
        self.assertEqual("Auto_HCL", self._hcl_elevation_max.state)

    def test_hand(self) -> None:
        """Test hand detection."""
        tests.helper.oh_item.item_state_change_event("Unittest_Color_min", 1000)
        tests.helper.oh_item.item_state_change_event("Unittest_Color_max", 1000)
        self.assertEqual("Auto_HCL", self._hcl_elevation_min.state)
        self.assertEqual("Auto_HCL", self._hcl_elevation_max.state)

        tests.helper.oh_item.item_state_change_event("Unittest_Color_min", 42)
        tests.helper.oh_item.item_state_change_event("Unittest_Color_max", 42)

        self.assertEqual("Hand", self._hcl_elevation_min.state)
        self.assertEqual("Hand", self._hcl_elevation_max.state)

    def test_focus(self) -> None:
        """Test focus."""
        self.assertEqual("Auto_HCL", self._hcl_elevation_min.state)
        self.assertEqual("Auto_HCL", self._hcl_elevation_max.state)

        tests.helper.oh_item.item_state_change_event("Unittest_Focus_max", "ON")
        self.assertEqual("Auto_HCL", self._hcl_elevation_min.state)
        self.assertEqual("Auto_Focus", self._hcl_elevation_max.state)

        tests.helper.oh_item.item_state_change_event("Unittest_Focus_max", "OFF")
        self.assertEqual("Auto_HCL", self._hcl_elevation_min.state)
        self.assertEqual("Auto_HCL", self._hcl_elevation_max.state)

    def test_sleep(self) -> None:
        """Test sleep."""
        self.assertEqual("Auto_HCL", self._hcl_elevation_min.state)
        self.assertEqual("Auto_HCL", self._hcl_elevation_max.state)

        tests.helper.oh_item.item_state_change_event("Unittest_Sleep_state", habapp_rules.system.SleepState.PRE_SLEEPING.value)
        self.assertEqual("Auto_HCL", self._hcl_elevation_min.state)
        self.assertEqual("Auto_Sleep_Active", self._hcl_elevation_max.state)

        tests.helper.oh_item.item_state_change_event("Unittest_Sleep_state", habapp_rules.system.SleepState.SLEEPING.value)
        self.assertEqual("Auto_HCL", self._hcl_elevation_min.state)
        self.assertEqual("Auto_Sleep_Active", self._hcl_elevation_max.state)

        tests.helper.oh_item.item_state_change_event("Unittest_Sleep_state", habapp_rules.system.SleepState.POST_SLEEPING.value)
        self.assertEqual("Auto_HCL", self._hcl_elevation_min.state)
        self.assertEqual("Auto_Sleep_Active", self._hcl_elevation_max.state)

        tests.helper.oh_item.item_state_change_event("Unittest_Sleep_state", habapp_rules.system.SleepState.AWAKE.value)
        self.assertEqual("Auto_HCL", self._hcl_elevation_min.state)
        self.assertEqual("Auto_Sleep_Post", self._hcl_elevation_max.state)

        self._hcl_elevation_max.post_sleep_timeout()
        self.assertEqual("Auto_HCL", self._hcl_elevation_max.state)

        # with focus on
        tests.helper.oh_item.item_state_change_event("Unittest_Focus_max", "ON")
        self.assertEqual("Auto_Focus", self._hcl_elevation_max.state)

        tests.helper.oh_item.item_state_change_event("Unittest_Sleep_state", habapp_rules.system.SleepState.PRE_SLEEPING.value)
        self.assertEqual("Auto_Sleep_Active", self._hcl_elevation_max.state)
        tests.helper.oh_item.assert_value("Unittest_Focus_max", "OFF")

    def test_switch_on(self) -> None:
        """Test switch on."""
        self._hcl_elevation_max.state = "Manual"
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Color_dimmer", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Manual_dimmer", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Color_dimmer_state", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Switch_on_dimmer", None)

        hcl_color_dimmer = habapp_rules.actors.light_hcl.HclElevation(
            habapp_rules.actors.config.light_hcl.HclElevationConfig(
                items=habapp_rules.actors.config.light_hcl.HclElevationItems(
                    color="Unittest_Color_dimmer",
                    manual="Unittest_Manual_dimmer",
                    elevation="Unittest_Elevation",
                    state="Unittest_Color_dimmer_state",
                    switch_on="Unittest_Switch_on_dimmer",
                )
            )
        )

        # event value == OFF
        with unittest.mock.patch("HABApp.rule.scheduler.job_builder.HABAppJobBuilder.once") as run_soon_mock:
            tests.helper.oh_item.item_state_change_event("Unittest_Switch_on_max", "OFF")
            tests.helper.oh_item.item_state_change_event("Unittest_Switch_on_dimmer", 0)
            run_soon_mock.assert_not_called()

        # state is not Auto_HCL
        with unittest.mock.patch("HABApp.rule.scheduler.job_builder.HABAppJobBuilder.once") as run_soon_mock:
            tests.helper.oh_item.item_state_change_event("Unittest_Switch_on_max", "ON")
            tests.helper.oh_item.item_state_change_event("Unittest_Switch_on_dimmer", 42)
            run_soon_mock.assert_not_called()

        # target_color is None
        self._hcl_elevation_max.state = "Auto_HCL"
        with (
            unittest.mock.patch("HABApp.rule.scheduler.job_builder.HABAppJobBuilder.once") as run_soon_mock,
            unittest.mock.patch.object(self._hcl_elevation_max, "_get_hcl_color", return_value=None),
            unittest.mock.patch.object(hcl_color_dimmer, "_get_hcl_color", return_value=None),
        ):
            tests.helper.oh_item.item_state_change_event("Unittest_Switch_on_max", "ON")
            tests.helper.oh_item.item_state_change_event("Unittest_Switch_on_dimmer", 43)
            run_soon_mock.assert_not_called()

        # target_color is a valid value
        self._hcl_elevation_max.state = "Auto_HCL"
        with (
            unittest.mock.patch("HABApp.rule.scheduler.job_builder.HABAppJobBuilder.once") as run_soon_mock,
            unittest.mock.patch.object(self._hcl_elevation_max, "_get_hcl_color", return_value=42),
            unittest.mock.patch.object(hcl_color_dimmer, "_get_hcl_color", return_value=44),
        ):
            tests.helper.oh_item.item_state_change_event("Unittest_Switch_on_max", "ON")
            tests.helper.oh_item.item_state_change_event("Unittest_Switch_on_dimmer", 80)

            run_soon_mock.assert_has_calls([
                unittest.mock.call(1, self._hcl_elevation_max._state_observer.send_command, 42),
                unittest.mock.call(1, hcl_color_dimmer._state_observer.send_command, 44),
            ])


class TestHclTime(tests.helper.test_case_base.TestCaseBaseStateMachine):
    """Tests for time-based HCL."""

    def setUp(self) -> None:
        """Set up tests."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Color_min", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Manual_min", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_Unittest_Color_min_state", None)

        self._config = habapp_rules.actors.config.light_hcl.HclTimeConfig(
            items=habapp_rules.actors.config.light_hcl.HclTimeItems(
                color=HABApp.openhab.items.NumberItem("Unittest_Color_min"),
                manual=HABApp.openhab.items.SwitchItem("Unittest_Manual_min"),
                state=HABApp.openhab.items.StringItem("H_Unittest_Color_min_state"),
            ),
            parameter=habapp_rules.actors.config.light_hcl.HclTimeParameter(
                color_map=[(2, 3000), (8, 4000), (12, 9000), (17, 9000), (20, 4000)],
            ),
        )

        self._rule = habapp_rules.actors.light_hcl.HclTime(self._config)

    def test_one_hour_later(self) -> None:
        """Test _one_hour_later."""
        TestCase = collections.namedtuple("TestCase", "configured, time, today_weekend_holiday, tomorrow_weekend_holiday, expected_result")

        test_cases = [
            # not configured -> always false
            TestCase(False, datetime.datetime(2023, 12, 19, 12), False, False, False),
            # 12:00 -> always false
            TestCase(True, datetime.datetime(2023, 12, 19, 12), False, False, False),
            TestCase(True, datetime.datetime(2023, 12, 19, 12), False, True, False),
            TestCase(True, datetime.datetime(2023, 12, 19, 12), True, False, False),
            TestCase(True, datetime.datetime(2023, 12, 19, 12), True, True, False),
            # 13:00 -> true if next day is a free day
            TestCase(True, datetime.datetime(2023, 12, 19, 13), False, False, False),
            TestCase(True, datetime.datetime(2023, 12, 19, 13), False, True, True),
            TestCase(True, datetime.datetime(2023, 12, 19, 13), True, False, False),
            TestCase(True, datetime.datetime(2023, 12, 19, 13), True, True, True),
            # 4:00 -> true if today is a free day
            TestCase(True, datetime.datetime(2023, 12, 19, 4), False, False, False),
            TestCase(True, datetime.datetime(2023, 12, 19, 4), False, True, True),
            TestCase(True, datetime.datetime(2023, 12, 19, 4), True, False, False),
            TestCase(True, datetime.datetime(2023, 12, 19, 4), True, True, True),
            # 5:00 -> always false
            TestCase(True, datetime.datetime(2023, 12, 19, 5), False, False, False),
            TestCase(True, datetime.datetime(2023, 12, 19, 5), False, True, False),
            TestCase(True, datetime.datetime(2023, 12, 19, 5), True, False, False),
            TestCase(True, datetime.datetime(2023, 12, 19, 5), True, True, False),
        ]

        with unittest.mock.patch("habapp_rules.core.type_of_day.is_holiday") as is_holiday_mock, unittest.mock.patch("habapp_rules.core.type_of_day.is_weekend") as is_weekend_mock:
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    # test holiday
                    is_holiday_mock.side_effect = [test_case.tomorrow_weekend_holiday, test_case.today_weekend_holiday]
                    is_weekend_mock.side_effect = [False, False]
                    self._rule._config.parameter.shift_weekend_holiday = test_case.configured

                    self.assertEqual(test_case.expected_result, self._rule._one_hour_later(test_case.time))

                    # test weekend
                    is_holiday_mock.side_effect = [False, False]
                    is_weekend_mock.side_effect = [test_case.tomorrow_weekend_holiday, test_case.today_weekend_holiday]
                    self._rule._config.parameter.shift_weekend_holiday = test_case.configured

                    self.assertEqual(test_case.expected_result, self._rule._one_hour_later(test_case.time))

    def test_get_hcl_color(self) -> None:
        """Test _get_hcl_color."""
        # test without color value as attribute
        TestCase = collections.namedtuple("TestCase", "test_time, output")

        test_cases = [
            TestCase(datetime.datetime(2023, 1, 1, 0, 0), 3333),
            TestCase(datetime.datetime(2023, 1, 1, 1, 0), 3167),
            TestCase(datetime.datetime(2023, 1, 1, 2, 0), 3000),
            TestCase(datetime.datetime(2023, 1, 1, 3, 30), 3250),
            TestCase(datetime.datetime(2023, 1, 1, 5, 0), 3500),
            TestCase(datetime.datetime(2023, 1, 1, 8, 0), 4000),
            TestCase(datetime.datetime(2023, 1, 1, 9, 0), 5250),
            TestCase(datetime.datetime(2023, 1, 1, 12, 0), 9000),
            TestCase(datetime.datetime(2023, 1, 1, 12, 10), 9000),
            TestCase(datetime.datetime(2023, 1, 1, 20, 0), 4000),
            TestCase(datetime.datetime(2023, 1, 1, 22, 0), 3667),
            TestCase(datetime.datetime(2023, 1, 1, 22, 12), 3633),
        ]

        with unittest.mock.patch("datetime.datetime") as datetime_mock:
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    datetime_mock.now.return_value = test_case.test_time
                    self.assertEqual(test_case.output, round(self._rule._get_hcl_color()))

        # test one hour later
        test_time = datetime.datetime(2023, 1, 1, 21, 0)
        with unittest.mock.patch.object(self._rule, "_one_hour_later", return_value=True), unittest.mock.patch("datetime.datetime") as datetime_mock:
            datetime_mock.now.return_value = test_time
            self.assertEqual(4000, round(self._rule._get_hcl_color()))

    def test_update_color(self) -> None:
        """Test _update_color."""
        with unittest.mock.patch.object(self._rule, "_get_hcl_color", return_value=42):
            self._rule._update_color()
        tests.helper.oh_item.assert_value("Unittest_Color_min", 42)

        # state is not Auto_HCL:
        tests.helper.oh_item.item_state_change_event("Unittest_Manual_min", "ON")
        with unittest.mock.patch.object(self._rule, "_get_hcl_color", return_value=123):
            self._rule._update_color()
        tests.helper.oh_item.assert_value("Unittest_Color_min", 42)



"""Test config models for monthly energy report rules."""

import unittest.mock

import HABApp
import pydantic

import habapp_rules.energy.config.monthly_report
import tests.helper.oh_item
import tests.helper.test_case_base


class TestEnergyShare(tests.helper.test_case_base.TestCaseBase):
    """Test EnergyShare dataclass."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Number_1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Number_2", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Switch_1", None)

    def test_init(self) -> None:
        """Test init."""
        # valid init
        energy_share = habapp_rules.energy.config.monthly_report.EnergyShare("Number_1", "First Number")
        self.assertEqual("Number_1", energy_share.energy_item.name)
        self.assertEqual("First Number", energy_share.chart_name)
        self.assertEqual(0, energy_share.monthly_power)

        expected_item = HABApp.openhab.items.NumberItem.get_item("Number_1")
        self.assertEqual(expected_item, energy_share.energy_item)

        # valid init with item
        energy_share = habapp_rules.energy.config.monthly_report.EnergyShare(expected_item, "First Number")
        self.assertEqual("Number_1", energy_share.energy_item.name)
        self.assertEqual("First Number", energy_share.chart_name)
        self.assertEqual(0, energy_share.monthly_power)

        # valid init with list of item names
        energy_share = habapp_rules.energy.config.monthly_report.EnergyShare(["Number_1", "Number_2"], "First Number")
        self.assertEqual(["Number_1", "Number_2"], [item.name for item in energy_share.energy_item])
        self.assertEqual("First Number", energy_share.chart_name)
        self.assertEqual(0, energy_share.monthly_power)

        # valid init with list of items
        energy_share = habapp_rules.energy.config.monthly_report.EnergyShare([expected_item, HABApp.openhab.items.NumberItem.get_item("Number_2")], "First Number")
        self.assertEqual(["Number_1", "Number_2"], [item.name for item in energy_share.energy_item])
        self.assertEqual("First Number", energy_share.chart_name)
        self.assertEqual(0, energy_share.monthly_power)

        # invalid init (Item not found)
        with self.assertRaises(pydantic.ValidationError):
            habapp_rules.energy.config.monthly_report.EnergyShare("Number_3", "Second Number")

        # invalid init (Item is not a number)
        with self.assertRaises(pydantic.ValidationError):
            habapp_rules.energy.config.monthly_report.EnergyShare("Switch_1", "Second Number")

    def test_get_energy_since(self) -> None:
        """Test get_energy_since."""
        # single item
        single = habapp_rules.energy.config.monthly_report.EnergyShare("Number_1", "First Number")
        tests.helper.oh_item.set_state("Number_1", 42)
        with unittest.mock.patch("habapp_rules.energy.helper.get_historic_value", side_effect=[12]) as get_historic_value_mock:
            time_mock = unittest.mock.MagicMock()
            self.assertEqual(30, single.get_energy_since(time_mock))
        get_historic_value_mock.assert_called_once_with(single.energy_item, time_mock)

        # multiple items
        multiple = habapp_rules.energy.config.monthly_report.EnergyShare(["Number_1", "Number_2"], "First Number")
        tests.helper.oh_item.set_state("Number_1", 100)
        tests.helper.oh_item.set_state("Number_2", 300)
        with unittest.mock.patch("habapp_rules.energy.helper.get_historic_value", side_effect=[50, 100]) as get_historic_value_mock:
            time_mock = unittest.mock.MagicMock()
            self.assertEqual(250, multiple.get_energy_since(time_mock))
        self.assertEqual(2, get_historic_value_mock.call_count)
        get_historic_value_mock.assert_has_calls([unittest.mock.call(multiple.energy_item[0], time_mock), unittest.mock.call(multiple.energy_item[1], time_mock)])

    def test_get_items_as_list(self) -> None:
        """Test get_items_as_list."""
        # single item
        single = habapp_rules.energy.config.monthly_report.EnergyShare("Number_1", "First Number")
        self.assertEqual(1, len(single.get_items_as_list))
        self.assertEqual("Number_1", single.get_items_as_list[0].name)

        # multiple items
        multiple = habapp_rules.energy.config.monthly_report.EnergyShare(["Number_1", "Number_2"], "First Number")
        self.assertEqual(2, len(multiple.get_items_as_list))
        self.assertEqual("Number_1", multiple.get_items_as_list[0].name)
        self.assertEqual("Number_2", multiple.get_items_as_list[1].name)



"""Unit-test for logic functions."""

import collections
import unittest.mock

import HABApp.openhab.items.switch_item

import habapp_rules.common.config.logic
import habapp_rules.common.logic
import habapp_rules.core.state_machine_rule
import tests.helper.oh_item
import tests.helper.test_case_base


class TestAndOR(tests.helper.test_case_base.TestCaseBase):
    """Tests for AND / OR."""

    def setUp(self) -> None:
        """Setup unit-tests."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        self.post_update_mock_patcher = unittest.mock.patch("HABApp.openhab.items.base_item.post_update", new=tests.helper.oh_item.send_command)
        self.addCleanup(self.post_update_mock_patcher.stop)
        self.post_update_mock_patcher.start()

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch_out", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch_in1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch_in2", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch_in3", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.ContactItem, "Unittest_Contact_out", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.ContactItem, "Unittest_Contact_in1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.ContactItem, "Unittest_Contact_in2", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.ContactItem, "Unittest_Contact_in3", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Number", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Dimmer", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_String", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.RollershutterItem, "Unittest_RollerShutter", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DatetimeItem, "Unittest_DateTime", None)

    def test_and_callback_switch(self) -> None:
        """Test <AND> for switch items."""
        TestStep = collections.namedtuple("TestStep", "event_item_name, event_item_value, expected_output")

        test_steps = [
            # test toggle of one switch
            TestStep("Unittest_Switch_in1", "ON", "OFF"),
            TestStep("Unittest_Switch_in1", "OFF", "OFF"),
            # switch on all
            TestStep("Unittest_Switch_in1", "ON", "OFF"),
            TestStep("Unittest_Switch_in2", "ON", "OFF"),
            TestStep("Unittest_Switch_in3", "ON", "ON"),
            # toggle one switch
            TestStep("Unittest_Switch_in1", "OFF", "OFF"),
            TestStep("Unittest_Switch_in1", "ON", "ON"),
            # switch off all
            TestStep("Unittest_Switch_in2", "OFF", "OFF"),
            TestStep("Unittest_Switch_in1", "OFF", "OFF"),
            TestStep("Unittest_Switch_in3", "OFF", "OFF"),
        ]

        config = habapp_rules.common.config.logic.BinaryLogicConfig(items=habapp_rules.common.config.logic.BinaryLogicItems(inputs=["Unittest_Switch_in1", "Unittest_Switch_in2", "Unittest_Switch_in3"], output="Unittest_Switch_out"))

        habapp_rules.common.logic.And(config)
        output_item = HABApp.openhab.items.SwitchItem.get_item("Unittest_Switch_out")

        for step in test_steps:
            tests.helper.oh_item.send_command(step.event_item_name, step.event_item_value)
            self.assertEqual(step.expected_output, output_item.value)

    def test_or_callback_switch(self) -> None:
        """Test <OR> for switch items."""
        TestStep = collections.namedtuple("TestStep", "event_item_name, event_item_value, expected_output")

        test_steps = [
            # test toggle of one switch
            TestStep("Unittest_Switch_in1", "ON", "ON"),
            TestStep("Unittest_Switch_in1", "OFF", "OFF"),
            # switch on all
            TestStep("Unittest_Switch_in1", "ON", "ON"),
            TestStep("Unittest_Switch_in2", "ON", "ON"),
            TestStep("Unittest_Switch_in3", "ON", "ON"),
            # toggle one switch
            TestStep("Unittest_Switch_in1", "OFF", "ON"),
            TestStep("Unittest_Switch_in1", "ON", "ON"),
            # switch off all
            TestStep("Unittest_Switch_in2", "OFF", "ON"),
            TestStep("Unittest_Switch_in1", "OFF", "ON"),
            TestStep("Unittest_Switch_in3", "OFF", "OFF"),
        ]

        config = habapp_rules.common.config.logic.BinaryLogicConfig(items=habapp_rules.common.config.logic.BinaryLogicItems(inputs=["Unittest_Switch_in1", "Unittest_Switch_in2", "Unittest_Switch_in3"], output="Unittest_Switch_out"))

        habapp_rules.common.logic.Or(config)
        output_item = HABApp.openhab.items.SwitchItem.get_item("Unittest_Switch_out")

        for step in test_steps:
            tests.helper.oh_item.send_command(step.event_item_name, step.event_item_value)
            self.assertEqual(step.expected_output, output_item.value)

    def test_and_callback_contact(self) -> None:
        """Test <AND> for contact items."""
        TestStep = collections.namedtuple("TestStep", "event_item_name, event_item_value, expected_output")

        test_steps = [
            # test toggle of one Contact
            TestStep("Unittest_Contact_in1", "CLOSED", "OPEN"),
            TestStep("Unittest_Contact_in1", "OPEN", "OPEN"),
            # Contact on all
            TestStep("Unittest_Contact_in1", "CLOSED", "OPEN"),
            TestStep("Unittest_Contact_in2", "CLOSED", "OPEN"),
            TestStep("Unittest_Contact_in3", "CLOSED", "CLOSED"),
            # toggle one Contact
            TestStep("Unittest_Contact_in1", "OPEN", "OPEN"),
            TestStep("Unittest_Contact_in1", "CLOSED", "CLOSED"),
            # Contact off all
            TestStep("Unittest_Contact_in2", "OPEN", "OPEN"),
            TestStep("Unittest_Contact_in1", "OPEN", "OPEN"),
            TestStep("Unittest_Contact_in3", "OPEN", "OPEN"),
        ]

        config = habapp_rules.common.config.logic.BinaryLogicConfig(items=habapp_rules.common.config.logic.BinaryLogicItems(inputs=["Unittest_Contact_in1", "Unittest_Contact_in2", "Unittest_Contact_in3"], output="Unittest_Contact_out"))

        habapp_rules.common.logic.And(config)
        output_item = HABApp.openhab.items.ContactItem.get_item("Unittest_Contact_out")

        for step in test_steps:
            tests.helper.oh_item.send_command(step.event_item_name, step.event_item_value)
            self.assertEqual(step.expected_output, output_item.value)

    def test_or_callback_contact(self) -> None:
        """Test <or> for contact items."""
        TestStep = collections.namedtuple("TestStep", "event_item_name, event_item_value, expected_output")

        test_steps = [
            # test toggle of one Contact
            TestStep("Unittest_Contact_in1", "CLOSED", "CLOSED"),
            TestStep("Unittest_Contact_in1", "OPEN", "OPEN"),
            # Contact on all
            TestStep("Unittest_Contact_in1", "CLOSED", "CLOSED"),
            TestStep("Unittest_Contact_in2", "CLOSED", "CLOSED"),
            TestStep("Unittest_Contact_in3", "CLOSED", "CLOSED"),
            # toggle one Contact
            TestStep("Unittest_Contact_in1", "OPEN", "CLOSED"),
            TestStep("Unittest_Contact_in1", "CLOSED", "CLOSED"),
            # Contact off all
            TestStep("Unittest_Contact_in2", "OPEN", "CLOSED"),
            TestStep("Unittest_Contact_in1", "OPEN", "CLOSED"),
            TestStep("Unittest_Contact_in3", "OPEN", "OPEN"),
        ]

        config = habapp_rules.common.config.logic.BinaryLogicConfig(items=habapp_rules.common.config.logic.BinaryLogicItems(inputs=["Unittest_Contact_in1", "Unittest_Contact_in2", "Unittest_Contact_in3"], output="Unittest_Contact_out"))

        habapp_rules.common.logic.Or(config)
        output_item = HABApp.openhab.items.ContactItem.get_item("Unittest_Contact_out")

        for step in test_steps:
            tests.helper.oh_item.send_command(step.event_item_name, step.event_item_value)
            self.assertEqual(step.expected_output, output_item.value)


class TestNumericLogic(tests.helper.test_case_base.TestCaseBase):
    """Tests Numeric logic rules."""

    def setUp(self) -> None:
        """Setup unit-tests."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Number_out_min", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Number_out_max", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Number_out_sum", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Number_in1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Number_in2", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Number_in3", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Dimmer_out_min", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Dimmer_out_max", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Dimmer_in1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Dimmer_in2", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Dimmer_in3", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.ContactItem, "Unittest_Contact", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_String", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.RollershutterItem, "Unittest_RollerShutter", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DatetimeItem, "Unittest_DateTime", None)

    def test_number_min_max_sum_without_filter(self) -> None:
        """Test min / max / sum for number items."""
        TestStep = collections.namedtuple("TestStep", "event_item_index, event_item_value, expected_min, expected_max, expected_sum")

        test_steps = [
            # test change single value
            TestStep(1, 100, 100, 100, 100),
            TestStep(1, 0, 0, 0, 0),
            TestStep(1, -100, -100, -100, -100),
            # change all values to 5000
            TestStep(1, 5000, 5000, 5000, 5000),
            TestStep(2, 5000, 5000, 5000, 10_000),
            TestStep(3, 5000, 5000, 5000, 15_000),
            # some random values
            TestStep(3, -1000, -1000, 5000, 9000),
            TestStep(3, -500, -500, 5000, 9500),
            TestStep(1, 200, -500, 5000, 4700),
        ]

        config_min = habapp_rules.common.config.logic.NumericLogicConfig(items=habapp_rules.common.config.logic.NumericLogicItems(inputs=["Unittest_Number_in1", "Unittest_Number_in2", "Unittest_Number_in3"], output="Unittest_Number_out_min"))

        config_max = habapp_rules.common.config.logic.NumericLogicConfig(items=habapp_rules.common.config.logic.NumericLogicItems(inputs=["Unittest_Number_in1", "Unittest_Number_in2", "Unittest_Number_in3"], output="Unittest_Number_out_max"))

        config_sum = habapp_rules.common.config.logic.NumericLogicConfig(items=habapp_rules.common.config.logic.NumericLogicItems(inputs=["Unittest_Number_in1", "Unittest_Number_in2", "Unittest_Number_in3"], output="Unittest_Number_out_sum"))

        habapp_rules.common.logic.Min(config_min)
        habapp_rules.common.logic.Max(config_max)
        habapp_rules.common.logic.Sum(config_sum)

        output_item_number_min = HABApp.openhab.items.NumberItem.get_item("Unittest_Number_out_min")
        output_item_number_max = HABApp.openhab.items.NumberItem.get_item("Unittest_Number_out_max")
        output_item_number_sum = HABApp.openhab.items.NumberItem.get_item("Unittest_Number_out_sum")

        for step in test_steps:
            tests.helper.oh_item.item_state_change_event(f"Unittest_Number_in{step.event_item_index}", step.event_item_value)

            self.assertEqual(step.expected_min, output_item_number_min.value)
            self.assertEqual(step.expected_max, output_item_number_max.value)
            self.assertEqual(step.expected_sum, output_item_number_sum.value)

    def test_dimmer_min_max_without_filter(self) -> None:
        """Test min / max for dimmer items."""
        TestStep = collections.namedtuple("TestStep", "event_item_index, event_item_value, expected_min, expected_max")

        test_steps = [
            # test change single value
            TestStep(1, 100, 100, 100),
            TestStep(1, 0, 0, 0),
            TestStep(1, 50, 50, 50),
            # change all values to 80
            TestStep(1, 80, 80, 80),
            TestStep(2, 80, 80, 80),
            TestStep(3, 80, 80, 80),
            # some random values
            TestStep(3, 1, 1, 80),
            TestStep(3, 20, 20, 80),
            TestStep(1, 50, 20, 80),
        ]

        config_min = habapp_rules.common.config.logic.NumericLogicConfig(items=habapp_rules.common.config.logic.NumericLogicItems(inputs=["Unittest_Dimmer_in1", "Unittest_Dimmer_in2", "Unittest_Dimmer_in3"], output="Unittest_Dimmer_out_min"))

        config_max = habapp_rules.common.config.logic.NumericLogicConfig(items=habapp_rules.common.config.logic.NumericLogicItems(inputs=["Unittest_Dimmer_in1", "Unittest_Dimmer_in2", "Unittest_Dimmer_in3"], output="Unittest_Dimmer_out_max"))

        habapp_rules.common.logic.Min(config_min)
        habapp_rules.common.logic.Max(config_max)

        output_item_dimmer_min = HABApp.openhab.items.DimmerItem.get_item("Unittest_Dimmer_out_min")
        output_item_dimmer_max = HABApp.openhab.items.DimmerItem.get_item("Unittest_Dimmer_out_max")

        for step in test_steps:
            tests.helper.oh_item.item_state_change_event(f"Unittest_Dimmer_in{step.event_item_index}", step.event_item_value)

            self.assertEqual(step.expected_min, output_item_dimmer_min.value)
            self.assertEqual(step.expected_max, output_item_dimmer_max.value)

    def test_cb_input_event(self) -> None:
        """Test _cb_input_event."""
        config_min = habapp_rules.common.config.logic.NumericLogicConfig(items=habapp_rules.common.config.logic.NumericLogicItems(inputs=["Unittest_Dimmer_in1", "Unittest_Dimmer_in2", "Unittest_Dimmer_in3"], output="Unittest_Dimmer_out_min"))

        config_max = habapp_rules.common.config.logic.NumericLogicConfig(items=habapp_rules.common.config.logic.NumericLogicItems(inputs=["Unittest_Dimmer_in1", "Unittest_Dimmer_in2", "Unittest_Dimmer_in3"], output="Unittest_Dimmer_out_max"))

        rule_min = habapp_rules.common.logic.Min(config_min)
        rule_max = habapp_rules.common.logic.Max(config_max)

        with unittest.mock.patch("habapp_rules.core.helper.filter_updated_items", return_value=[None]), unittest.mock.patch.object(rule_min, "_set_output_state") as set_output_mock:
            rule_min._cb_input_event(None)
        set_output_mock.assert_not_called()

        with unittest.mock.patch("habapp_rules.core.helper.filter_updated_items", return_value=[None]), unittest.mock.patch.object(rule_max, "_set_output_state") as set_output_mock:
            rule_max._cb_input_event(None)
        set_output_mock.assert_not_called()

    def test_exception_dimmer_sum(self) -> None:
        """Test exception if Sum is instantiated with dimmer items."""
        config_max = habapp_rules.common.config.logic.NumericLogicConfig(items=habapp_rules.common.config.logic.NumericLogicItems(inputs=["Unittest_Dimmer_in1", "Unittest_Dimmer_in2", "Unittest_Dimmer_in3"], output="Unittest_Dimmer_out_max"))

        with self.assertRaises(TypeError):
            habapp_rules.common.logic.Sum(config_max)


class TestInvertValue(tests.helper.test_case_base.TestCaseBase):
    """Tests InvertValue rule."""

    def setUp(self) -> None:
        """Setup unit-tests."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Input", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Output", None)

    def test_invert_value_without_pos_neg(self) -> None:
        """Test invert value rule without pos / neg set."""
        TestCase = collections.namedtuple("TestCase", "input, expected_output")

        test_cases = [TestCase(10, -10), TestCase(1, -1), TestCase(0.1, -0.1), TestCase(0, 0), TestCase(-0.1, 0.1), TestCase(-1, 1), TestCase(-10, 10)]

        config = habapp_rules.common.config.logic.InvertValueConfig(items=habapp_rules.common.config.logic.InvertValueItems(input="Unittest_Input", output="Unittest_Output"))

        habapp_rules.common.logic.InvertValue(config)

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.item_state_change_event("Unittest_Input", test_case.input)
                tests.helper.oh_item.assert_value("Unittest_Output", test_case.expected_output)

    def test_invert_value_with_only_pos(self) -> None:
        """Test invert value rule with only pos is set."""
        TestCase = collections.namedtuple("TestCase", "input, expected_output")

        test_cases = [TestCase(10, 0), TestCase(1, 0), TestCase(0.1, 0), TestCase(0, 0), TestCase(-0.1, 0.1), TestCase(-1, 1), TestCase(-10, 10)]

        config = habapp_rules.common.config.logic.InvertValueConfig(
            items=habapp_rules.common.config.logic.InvertValueItems(input="Unittest_Input", output="Unittest_Output"), parameter=habapp_rules.common.config.logic.InvertValueParameter(only_positive=True)
        )

        habapp_rules.common.logic.InvertValue(config)

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.item_state_change_event("Unittest_Input", test_case.input)
                tests.helper.oh_item.assert_value("Unittest_Output", test_case.expected_output)

    def test_invert_value_with_only_neg(self) -> None:
        """Test invert value rule with only neg is set."""
        TestCase = collections.namedtuple("TestCase", "input, expected_output")

        test_cases = [TestCase(10, -10), TestCase(1, -1), TestCase(0.1, -0.1), TestCase(0, 0), TestCase(-0.1, 0), TestCase(-1, 0), TestCase(-10, 0)]

        config = habapp_rules.common.config.logic.InvertValueConfig(
            items=habapp_rules.common.config.logic.InvertValueItems(input="Unittest_Input", output="Unittest_Output"), parameter=habapp_rules.common.config.logic.InvertValueParameter(only_negative=True)
        )

        habapp_rules.common.logic.InvertValue(config)

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.item_state_change_event("Unittest_Input", test_case.input)
                tests.helper.oh_item.assert_value("Unittest_Output", test_case.expected_output)



"""Test sleep rule."""

import collections
import datetime
import pathlib
import sys
import unittest
import unittest.mock

import HABApp.rule.rule

import habapp_rules.core.state_machine_rule
import habapp_rules.system.config.sleep
import habapp_rules.system.sleep
import tests.helper.graph_machines
import tests.helper.oh_item
import tests.helper.test_case_base
import tests.helper.timer


class TestSleep(tests.helper.test_case_base.TestCaseBaseStateMachine):
    """Tests cases for testing presence rule."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBaseStateMachine.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Sleep", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Sleep_Request", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Lock", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Lock_Request", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Display_Text", "")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_Sleep_Unittest_Sleep_state", "")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "CustomState", "")

        config = habapp_rules.system.config.sleep.SleepConfig(
            items=habapp_rules.system.config.sleep.SleepItems(sleep="Unittest_Sleep", sleep_request="Unittest_Sleep_Request", lock="Unittest_Lock", lock_request="Unittest_Lock_Request", display_text="Unittest_Display_Text", state="CustomState")
        )

        self._sleep = habapp_rules.system.sleep.Sleep(config)

    def test_init_with_none(self) -> None:
        """Test __init__ with None values."""
        tests.helper.oh_item.set_state("Unittest_Sleep", None)
        tests.helper.oh_item.set_state("Unittest_Sleep_Request", None)
        tests.helper.oh_item.set_state("Unittest_Lock", None)
        tests.helper.oh_item.set_state("Unittest_Lock_Request", None)
        tests.helper.oh_item.set_state("Unittest_Display_Text", None)
        tests.helper.oh_item.set_state("CustomState", None)

        config = habapp_rules.system.config.sleep.SleepConfig(
            items=habapp_rules.system.config.sleep.SleepItems(sleep="Unittest_Sleep", sleep_request="Unittest_Sleep_Request", lock="Unittest_Lock", lock_request="Unittest_Lock_Request", display_text="Unittest_Display_Text", state="CustomState")
        )

        habapp_rules.system.sleep.Sleep(config)

    @unittest.skipIf(sys.platform != "win32", "Should only run on windows when graphviz is installed")
    def test_create_graph(self) -> None:  # pragma: no cover
        """Create state machine graph for documentation."""
        presence_graph = tests.helper.graph_machines.GraphMachineTimer(model=self._sleep, states=self._sleep.states, transitions=self._sleep.trans, initial=self._sleep.state, show_conditions=True)

        picture_dir = pathlib.Path(__file__).parent / "_state_charts" / "Sleep"
        if not picture_dir.is_dir():
            picture_dir.mkdir(parents=True)
        presence_graph.get_graph().draw(picture_dir / "Sleep.png", format="png", prog="dot")

    def test_enums(self) -> None:
        """Test if all enums from __init__.py are implemented."""
        implemented_states = list(self._sleep.state_machine.states)
        enum_states = [state.value for state in habapp_rules.system.SleepState] + ["initial"]
        self.assertEqual(len(enum_states), len(implemented_states))
        self.assertTrue(all(state in enum_states for state in implemented_states))

    def test__init__(self) -> None:
        """Test init of sleep class."""
        TestCase = collections.namedtuple("TestCase", "sleep_request_state, lock_request_state, lock_state")

        test_cases = [
            TestCase("OFF", "OFF", "OFF"),
            TestCase("OFF", "ON", "ON"),
            TestCase("ON", "OFF", "OFF"),
            TestCase("ON", "ON", "OFF"),
        ]

        config = habapp_rules.system.config.sleep.SleepConfig(
            items=habapp_rules.system.config.sleep.SleepItems(sleep="Unittest_Sleep", sleep_request="Unittest_Sleep_Request", lock="Unittest_Lock", lock_request="Unittest_Lock_Request", display_text="Unittest_Display_Text", state="CustomState")
        )

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.set_state("Unittest_Sleep_Request", test_case.sleep_request_state)
                tests.helper.oh_item.set_state("Unittest_Lock_Request", test_case.lock_request_state)

                sleep = habapp_rules.system.sleep.Sleep(config)

                self.assertEqual(sleep.sleep_request_active, test_case.sleep_request_state == "ON", test_case)
                self.assertEqual(sleep.lock_request_active, test_case.lock_request_state == "ON", test_case)
                tests.helper.oh_item.assert_value("Unittest_Sleep", test_case.sleep_request_state)
                tests.helper.oh_item.assert_value("Unittest_Lock", test_case.lock_state)

    def test_get_initial_state(self) -> None:
        """Test getting initial state."""
        TestCase = collections.namedtuple("TestCase", "sleep_request, lock_request, expected_state")

        test_cases = [
            TestCase(sleep_request="OFF", lock_request="OFF", expected_state="awake"),
            TestCase(sleep_request="OFF", lock_request="ON", expected_state="locked"),
            TestCase(sleep_request="ON", lock_request="OFF", expected_state="sleeping"),
            TestCase(sleep_request="ON", lock_request="ON", expected_state="sleeping"),
            TestCase(sleep_request=None, lock_request="ON", expected_state="locked"),
            TestCase(sleep_request=None, lock_request="OFF", expected_state="default"),
            TestCase(sleep_request="ON", lock_request=None, expected_state="sleeping"),
            TestCase(sleep_request="OFF", lock_request=None, expected_state="awake"),
            TestCase(sleep_request=None, lock_request=None, expected_state="default"),
        ]

        for test_case in test_cases:
            tests.helper.oh_item.set_state("Unittest_Sleep_Request", test_case.sleep_request)
            tests.helper.oh_item.set_state("Unittest_Lock_Request", test_case.lock_request)

            self.assertEqual(self._sleep._get_initial_state("default"), test_case.expected_state, test_case)

    def test__get_display_text(self) -> None:
        """Test getting display text."""
        TestCase = collections.namedtuple("TestCase", "state, text")
        test_cases = [TestCase("awake", "Schlafen"), TestCase("pre_sleeping", "Guten Schlaf"), TestCase("sleeping", "Aufstehen"), TestCase("post_sleeping", "Guten Morgen"), TestCase("locked", "Gesperrt"), TestCase(None, "")]

        for test_case in test_cases:
            self._sleep.state = test_case.state
            self.assertEqual(test_case.text, self._sleep._Sleep__get_display_text())

    def test_normal_cycle_all_items(self) -> None:
        """Test normal behavior with all items available."""
        # check initial state
        tests.helper.oh_item.assert_value("CustomState", "awake")

        # start sleeping
        tests.helper.oh_item.send_command("Unittest_Sleep_Request", "ON", "OFF")
        self.assertEqual(self._sleep.state, "pre_sleeping")
        tests.helper.oh_item.assert_value("CustomState", "pre_sleeping")
        tests.helper.oh_item.assert_value("Unittest_Sleep", "ON")
        tests.helper.oh_item.assert_value("Unittest_Lock", "ON")
        tests.helper.oh_item.assert_value("Unittest_Display_Text", "Guten Schlaf")
        self.transitions_timer_mock.assert_called_with(3, unittest.mock.ANY, args=unittest.mock.ANY)

        # pre_sleeping timeout -> sleep
        tests.helper.timer.call_timeout(self.transitions_timer_mock)
        self.assertEqual(self._sleep.state, "sleeping")
        tests.helper.oh_item.assert_value("CustomState", "sleeping")
        tests.helper.oh_item.assert_value("Unittest_Sleep", "ON")
        tests.helper.oh_item.assert_value("Unittest_Lock", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Display_Text", "Aufstehen")

        # stop sleeping
        self.transitions_timer_mock.reset_mock()
        tests.helper.oh_item.send_command("Unittest_Sleep_Request", "OFF", "ON")
        self.assertEqual(self._sleep.state, "post_sleeping")
        tests.helper.oh_item.assert_value("CustomState", "post_sleeping")
        tests.helper.oh_item.assert_value("Unittest_Sleep", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Lock", "ON")
        tests.helper.oh_item.assert_value("Unittest_Display_Text", "Guten Morgen")
        self.transitions_timer_mock.assert_called_with(3, unittest.mock.ANY, args=unittest.mock.ANY)

        # post_sleeping check if sleep change is ignored
        tests.helper.oh_item.send_command("Unittest_Sleep_Request", "ON", "OFF")
        self.assertEqual(self._sleep.state, "post_sleeping")

        # post_sleeping timeout -> awake
        tests.helper.timer.call_timeout(self.transitions_timer_mock)
        self.assertEqual(self._sleep.state, "awake")
        tests.helper.oh_item.assert_value("CustomState", "awake")
        tests.helper.oh_item.assert_value("Unittest_Sleep", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Lock", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Display_Text", "Schlafen")

    def test_lock_transitions(self) -> None:
        """Test all transitions from and to locked state."""
        # check correct initial state
        tests.helper.oh_item.assert_value("CustomState", "awake")
        tests.helper.oh_item.assert_value("Unittest_Lock", "OFF")

        # set lock_request. expected: locked state, lock active, sleep off
        tests.helper.oh_item.send_command("Unittest_Lock_Request", "ON", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Lock", "ON")
        tests.helper.oh_item.assert_value("CustomState", "locked")
        tests.helper.oh_item.assert_value("Unittest_Sleep", "OFF")

        # release lock and come back to awake state
        tests.helper.oh_item.send_command("Unittest_Lock_Request", "OFF", "ON")
        tests.helper.oh_item.assert_value("Unittest_Lock", "OFF")
        tests.helper.oh_item.assert_value("CustomState", "awake")
        tests.helper.oh_item.assert_value("Unittest_Sleep", "OFF")

        # set lock_request and shortly after send sleep request -> locked expected
        tests.helper.oh_item.send_command("Unittest_Lock_Request", "ON", "OFF")
        tests.helper.oh_item.send_command("Unittest_Sleep_Request", "ON", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Sleep_Request", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Lock", "ON")
        tests.helper.oh_item.assert_value("CustomState", "locked")
        tests.helper.oh_item.assert_value("Unittest_Sleep", "OFF")

        # release lock and jump back to awake
        tests.helper.oh_item.send_command("Unittest_Lock_Request", "OFF", "ON")
        tests.helper.oh_item.assert_value("Unittest_Lock", "OFF")
        tests.helper.oh_item.assert_value("CustomState", "awake")
        tests.helper.oh_item.assert_value("Unittest_Sleep", "OFF")

        # start sleeping
        tests.helper.oh_item.send_command("Unittest_Sleep_Request", "ON", "OFF")
        tests.helper.oh_item.assert_value("CustomState", "pre_sleeping")

        # activate lock, remove sleep request and wait all timer -> expected state == locked
        tests.helper.oh_item.send_command("Unittest_Lock_Request", "ON", "OFF")
        tests.helper.oh_item.assert_value("CustomState", "pre_sleeping")
        tests.helper.timer.call_timeout(self.transitions_timer_mock)
        tests.helper.oh_item.assert_value("CustomState", "sleeping")
        tests.helper.oh_item.send_command("Unittest_Sleep_Request", "OFF", "ON")
        tests.helper.oh_item.assert_value("CustomState", "post_sleeping")
        tests.helper.timer.call_timeout(self.transitions_timer_mock)
        tests.helper.oh_item.assert_value("CustomState", "locked")

        # go back to pre_sleeping and check lock + end sleep in pre_sleeping -> expected state = locked
        tests.helper.oh_item.send_command("Unittest_Sleep_Request", "ON", "OFF")
        tests.helper.oh_item.assert_value("CustomState", "locked")
        tests.helper.oh_item.send_command("Unittest_Lock_Request", "OFF", "ON")
        tests.helper.oh_item.assert_value("CustomState", "awake")
        tests.helper.oh_item.send_command("Unittest_Lock_Request", "ON", "OFF")
        tests.helper.oh_item.assert_value("CustomState", "locked")

    def test_minimal_items(self) -> None:
        """Test Sleeping class with minimal set of items."""
        # delete sleep rule from init
        self._runner.loaded_rules[0]._habapp_ctx.unload_rule()
        self._runner.loaded_rules.clear()

        tests.helper.oh_item.remove_mocked_item_by_name("Unittest_Lock")
        tests.helper.oh_item.remove_mocked_item_by_name("Unittest_Lock_Request")
        tests.helper.oh_item.remove_mocked_item_by_name("Unittest_Display_Text")

        config = habapp_rules.system.config.sleep.SleepConfig(
            items=habapp_rules.system.config.sleep.SleepItems(
                sleep="Unittest_Sleep",
                sleep_request="Unittest_Sleep_Request",
                state="H_Sleep_Unittest_Sleep_state",
            )
        )

        sleep = habapp_rules.system.sleep.Sleep(config)

        self.assertIsNone(sleep._config.items.display_text)
        self.assertIsNone(sleep._config.items.lock)
        self.assertIsNone(sleep._config.items.lock_request)

        # check initial state
        tests.helper.oh_item.assert_value("CustomState", "awake")

        # start sleeping
        tests.helper.oh_item.send_command("Unittest_Sleep_Request", "ON", "OFF")
        self.assertEqual(sleep.state, "pre_sleeping")
        tests.helper.oh_item.assert_value("H_Sleep_Unittest_Sleep_state", "pre_sleeping")
        tests.helper.oh_item.assert_value("Unittest_Sleep", "ON")
        self.transitions_timer_mock.assert_called_with(3, unittest.mock.ANY, args=unittest.mock.ANY)

        # pre_sleeping timeout -> sleep
        tests.helper.timer.call_timeout(self.transitions_timer_mock)
        self.assertEqual(sleep.state, "sleeping")
        tests.helper.oh_item.assert_value("H_Sleep_Unittest_Sleep_state", "sleeping")
        tests.helper.oh_item.assert_value("Unittest_Sleep", "ON")

        # stop sleeping
        self.transitions_timer_mock.reset_mock()
        tests.helper.oh_item.send_command("Unittest_Sleep_Request", "OFF", "ON")
        self.assertEqual(sleep.state, "post_sleeping")
        tests.helper.oh_item.assert_value("H_Sleep_Unittest_Sleep_state", "post_sleeping")
        tests.helper.oh_item.assert_value("Unittest_Sleep", "OFF")
        self.transitions_timer_mock.assert_called_with(3, unittest.mock.ANY, args=unittest.mock.ANY)

        # post_sleeping timeout -> awake
        tests.helper.timer.call_timeout(self.transitions_timer_mock)
        self.assertEqual(sleep.state, "awake")
        tests.helper.oh_item.assert_value("H_Sleep_Unittest_Sleep_state", "awake")
        tests.helper.oh_item.assert_value("Unittest_Sleep", "OFF")


class TestLinkSleep(tests.helper.test_case_base.TestCaseBase):
    """Test LinkSleep."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Sleep1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Sleep2_req", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Sleep3_req", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Sleep4", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Sleep5_req", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Sleep6_req", None)

        config_full_day = habapp_rules.system.config.sleep.LinkSleepConfig(
            items=habapp_rules.system.config.sleep.LinkSleepItems(
                sleep_master="Unittest_Sleep1",
                sleep_request_slaves=["Unittest_Sleep2_req", "Unittest_Sleep3_req"],
            )
        )

        config_night = habapp_rules.system.config.sleep.LinkSleepConfig(
            items=habapp_rules.system.config.sleep.LinkSleepItems(
                sleep_master="Unittest_Sleep4",
                sleep_request_slaves=["Unittest_Sleep5_req", "Unittest_Sleep6_req"],
            ),
            parameter=habapp_rules.system.config.sleep.LinkSleepParameter(
                start_time=datetime.time(22),
                end_time=datetime.time(10),
            ),
        )

        self._link_full_day = habapp_rules.system.sleep.LinkSleep(config_full_day)
        self._link_night = habapp_rules.system.sleep.LinkSleep(config_night)

    def test_init_with_feedback(self) -> None:
        """Test init with feedback item."""
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Link_Active", None)
        config = habapp_rules.system.config.sleep.LinkSleepConfig(
            items=habapp_rules.system.config.sleep.LinkSleepItems(
                sleep_master="Unittest_Sleep1",
                sleep_request_slaves=["Unittest_Sleep2_req", "Unittest_Sleep3_req"],
                link_active_feedback="Unittest_Link_Active",
            )
        )

        rule = habapp_rules.system.sleep.LinkSleep(config)

        self.assertEqual("Unittest_Link_Active", rule._config.items.link_active_feedback.name)

    def test_check_time_in_window(self) -> None:
        """Test check_time_in_window."""
        TestCase = collections.namedtuple("TestCase", "start, end, now, expected_result")

        test_cases = [
            # full day
            TestCase(datetime.time(0), datetime.time(23, 59), datetime.time(0, 0), True),
            TestCase(datetime.time(0), datetime.time(23, 59), datetime.time(12), True),
            TestCase(datetime.time(0), datetime.time(23, 59), datetime.time(23, 59), True),
            # range during day
            TestCase(datetime.time(10), datetime.time(16), datetime.time(0, 0), False),
            TestCase(datetime.time(10), datetime.time(16), datetime.time(9, 59), False),
            TestCase(datetime.time(10), datetime.time(16), datetime.time(10), True),
            TestCase(datetime.time(10), datetime.time(16), datetime.time(10, 1), True),
            TestCase(datetime.time(10), datetime.time(16), datetime.time(15, 59), True),
            TestCase(datetime.time(10), datetime.time(16), datetime.time(16), True),
            TestCase(datetime.time(10), datetime.time(16), datetime.time(16, 1), False),
            TestCase(datetime.time(10), datetime.time(16), datetime.time(23, 59), False),
            # range over midnight day
            TestCase(datetime.time(22), datetime.time(5), datetime.time(12), False),
            TestCase(datetime.time(22), datetime.time(5), datetime.time(21, 59), False),
            TestCase(datetime.time(22), datetime.time(5), datetime.time(22), True),
            TestCase(datetime.time(22), datetime.time(5), datetime.time(22, 1), True),
            TestCase(datetime.time(22), datetime.time(5), datetime.time(0), True),
            TestCase(datetime.time(22), datetime.time(5), datetime.time(4, 59), True),
            TestCase(datetime.time(22), datetime.time(5), datetime.time(5), True),
            TestCase(datetime.time(22), datetime.time(5), datetime.time(5, 1), False),
        ]

        with unittest.mock.patch("datetime.datetime") as datetime_mock:
            now_mock = unittest.mock.MagicMock()
            datetime_mock.now.return_value = now_mock
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    now_mock.time.return_value = test_case.now

                    self._link_full_day._config.parameter.link_time_start = test_case.start
                    self._link_full_day._config.parameter.link_time_end = test_case.end

                    self.assertEqual(test_case.expected_result, self._link_full_day._check_time_in_window())

    def test_cb_master(self) -> None:
        """Test _cb_master."""
        # during active time
        with unittest.mock.patch.object(self._link_full_day, "_check_time_in_window", return_value=True):
            tests.helper.oh_item.assert_value("Unittest_Sleep2_req", None)
            tests.helper.oh_item.assert_value("Unittest_Sleep3_req", None)

            tests.helper.oh_item.item_state_change_event("Unittest_Sleep1", "ON")
            tests.helper.oh_item.assert_value("Unittest_Sleep2_req", "ON")
            tests.helper.oh_item.assert_value("Unittest_Sleep3_req", "ON")

        # during inactive time
        with unittest.mock.patch.object(self._link_night, "_check_time_in_window", return_value=False):
            tests.helper.oh_item.assert_value("Unittest_Sleep5_req", None)
            tests.helper.oh_item.assert_value("Unittest_Sleep6_req", None)

            tests.helper.oh_item.item_state_change_event("Unittest_Sleep4", "ON")
            tests.helper.oh_item.assert_value("Unittest_Sleep5_req", None)
            tests.helper.oh_item.assert_value("Unittest_Sleep6_req", None)

    def test_set_link_active_feedback(self) -> None:
        """Test _set_link_active_feedback."""
        with unittest.mock.patch.object(self._link_full_day._config.items, "link_active_feedback") as item_link_active_mock:
            self._link_full_day._set_link_active_feedback("ON")
        item_link_active_mock.oh_send_command.assert_called_once_with("ON")

        with unittest.mock.patch.object(self._link_full_day._config.items, "link_active_feedback") as item_link_active_mock:
            self._link_full_day._set_link_active_feedback("OFF")
        item_link_active_mock.oh_send_command.assert_called_once_with("OFF")



"""States of all system state machines."""

import enum


class SleepState(enum.Enum):
    """Sleep states."""

    AWAKE = "awake"
    PRE_SLEEPING = "pre_sleeping"
    SLEEPING = "sleeping"
    POST_SLEEPING = "post_sleeping"
    LOCKED = "locked"


class PresenceState(enum.Enum):
    """Presence states."""

    PRESENCE = "presence"
    LEAVING = "leaving"
    ABSENCE = "absence"
    LONG_ABSENCE = "long_absence"



import unittest.mock

import HABApp

import habapp_rules.network.wol
import tests.helper.oh_item
import tests.helper.test_case_base
from habapp_rules.network.config.wol import WolConfig, WolItems, WolParameter


class TestWOL(tests.helper.test_case_base.TestCaseBase):
    """Test Wol rule."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_WOL_min", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_WOL_max", None)

        self._config_min = WolConfig(items=WolItems(trigger_wol="Unittest_WOL_min"), parameter=WolParameter(mac_address="12:34:56:78:9a:ff"))
        self._config_max = WolConfig(items=WolItems(trigger_wol="Unittest_WOL_max"), parameter=WolParameter(mac_address="ab:cd:56:78:9a:ff", friendly_name="Some better name"))

        self._rule_min = habapp_rules.network.wol.Wol(self._config_min)
        self._rule_max = habapp_rules.network.wol.Wol(self._config_max)

    def test_trigger(self) -> None:
        """Test trigger of WOL."""
        # min
        with unittest.mock.patch("habapp_rules.network.wol.send_magic_packet") as send_magic_packet_mock, unittest.mock.patch("habapp_rules.network.wol.LOGGER") as logger_mock:
            tests.helper.oh_item.item_state_change_event("Unittest_WOL_min", "ON")
            send_magic_packet_mock.assert_called_once_with("12:34:56:78:9a:ff")
            logger_mock.info.assert_called_once_with("Triggered WOL for '12:34:56:78:9a:ff'")
            tests.helper.oh_item.assert_value("Unittest_WOL_min", "OFF")

        # max
        with unittest.mock.patch("habapp_rules.network.wol.send_magic_packet") as send_magic_packet_mock, unittest.mock.patch("habapp_rules.network.wol.LOGGER") as logger_mock:
            tests.helper.oh_item.item_state_change_event("Unittest_WOL_max", "ON")
            send_magic_packet_mock.assert_called_once_with("ab:cd:56:78:9a:ff")
            logger_mock.info.assert_called_once_with("Triggered WOL for 'Some better name'")
            tests.helper.oh_item.assert_value("Unittest_WOL_max", "OFF")



"""Tests for motion sensors."""

import collections
import pathlib
import sys
import unittest
import unittest.mock

import HABApp

import habapp_rules.core.exceptions
import habapp_rules.core.state_machine_rule
import habapp_rules.sensors.config.humidity
import habapp_rules.sensors.humidity
import habapp_rules.system
import tests.helper.graph_machines
import tests.helper.oh_item
import tests.helper.test_case_base
import tests.helper.timer


class TestMotion(tests.helper.test_case_base.TestCaseBaseStateMachine):
    """Tests cases for testing motion sensors rule."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBaseStateMachine.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Humidity", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Output", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_Unittest_Output_state", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Custom_Name", None)

        config = habapp_rules.sensors.config.humidity.HumiditySwitchConfig(items=habapp_rules.sensors.config.humidity.HumiditySwitchItems(humidity="Unittest_Humidity", output="Unittest_Output", state="H_Unittest_Output_state"))

        self.humidity = habapp_rules.sensors.humidity.HumiditySwitch(config)

    @unittest.skipIf(sys.platform != "win32", "Should only run on windows when graphviz is installed")
    def test_create_graph(self) -> None:  # pragma: no cover
        """Create state machine graph for documentation."""
        picture_dir = pathlib.Path(__file__).parent / "_state_charts" / "Humidity"
        if not picture_dir.is_dir():
            picture_dir.mkdir(parents=True)

        humidity_graph = tests.helper.graph_machines.HierarchicalGraphMachineTimer(model=tests.helper.graph_machines.FakeModel(), states=self.humidity.states, transitions=self.humidity.trans, initial=self.humidity.state, show_conditions=True)

        humidity_graph.get_graph().draw(picture_dir / "Humidity.png", format="png", prog="dot")

    def test_init(self) -> None:
        """Test init."""
        full_config = habapp_rules.sensors.config.humidity.HumiditySwitchConfig(
            items=habapp_rules.sensors.config.humidity.HumiditySwitchItems(humidity="Unittest_Humidity", output="Unittest_Output", state="Custom_Name"),
            parameter=habapp_rules.sensors.config.humidity.HumiditySwitchParameter(absolute_threshold=70, extended_time=42),
        )

        humidity = habapp_rules.sensors.humidity.HumiditySwitch(full_config)
        self.assertEqual(70, humidity._config.parameter.absolute_threshold)
        self.assertEqual(42, humidity.state_machine.get_state("on_Extended").timeout)
        self.assertEqual("Custom_Name", humidity._item_state.name)

    def test_get_initial_state(self) -> None:
        """Test get_initial_state."""
        TestCase = collections.namedtuple("TestCase", "humidity_value, expected_state")

        test_cases = [
            TestCase(None, "off"),
            TestCase(64, "off"),
            TestCase(65, "on"),
            TestCase(66, "on"),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.set_state("Unittest_Humidity", test_case.humidity_value)
                self.assertEqual(test_case.expected_state, self.humidity._get_initial_state())

    def test_check_high_humidity(self) -> None:
        """Test check_high_humidity."""
        TestCase = collections.namedtuple("TestCase", "item_value,given_value, expected_result")

        test_cases = [
            # False | False -> False
            TestCase(None, None, False),
            TestCase(None, 64, False),
            TestCase(64, None, False),
            TestCase(64, 64, False),
            # False | True -> True
            TestCase(None, 65, True),
            TestCase(64, 65, True),
            # True | False -> False
            TestCase(65, None, True),
            TestCase(65, 64, False),
            # True | True -> True
            TestCase(65, 65, True),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.set_state("Unittest_Humidity", test_case.item_value)
                self.assertEqual(test_case.expected_result, self.humidity._check_high_humidity(test_case.given_value))

    def test_cb_humidity(self) -> None:
        """Test _cb_humidity."""
        with (
            unittest.mock.patch.object(self.humidity, "high_humidity_start") as start_mock,
            unittest.mock.patch.object(self.humidity, "high_humidity_end") as end_mock,
            unittest.mock.patch.object(self.humidity, "_check_high_humidity", return_value=True) as check_mock,
        ):
            tests.helper.oh_item.item_state_event("Unittest_Humidity", 99)
            check_mock.assert_called_once_with(99)
            start_mock.assert_called_once()
            end_mock.assert_not_called()

        with (
            unittest.mock.patch.object(self.humidity, "high_humidity_start") as start_mock,
            unittest.mock.patch.object(self.humidity, "high_humidity_end") as end_mock,
            unittest.mock.patch.object(self.humidity, "_check_high_humidity", return_value=False) as check_mock,
        ):
            tests.helper.oh_item.item_state_event("Unittest_Humidity", 42)
            check_mock.assert_called_once_with(42)
            start_mock.assert_not_called()
            end_mock.assert_called_once()

    def test_states(self) -> None:
        """Test states."""
        self.assertEqual("off", self.humidity.state)

        # some humidity changes below threshold
        tests.helper.oh_item.item_state_event("Unittest_Humidity", 64)
        self.assertEqual("off", self.humidity.state)
        tests.helper.oh_item.assert_value("Unittest_Output", "OFF")
        tests.helper.oh_item.item_state_event("Unittest_Humidity", 10)
        self.assertEqual("off", self.humidity.state)
        tests.helper.oh_item.assert_value("Unittest_Output", "OFF")

        # some humidity changes above threshold
        tests.helper.oh_item.item_state_event("Unittest_Humidity", 65)
        self.assertEqual("on_HighHumidity", self.humidity.state)
        tests.helper.oh_item.assert_value("Unittest_Output", "ON")
        tests.helper.oh_item.item_state_event("Unittest_Humidity", 100)
        self.assertEqual("on_HighHumidity", self.humidity.state)
        tests.helper.oh_item.assert_value("Unittest_Output", "ON")

        # humidity below threshold again
        tests.helper.oh_item.item_state_event("Unittest_Humidity", 50)
        self.assertEqual("on_Extended", self.humidity.state)
        tests.helper.oh_item.assert_value("Unittest_Output", "ON")

        # humidity above threshold again
        tests.helper.oh_item.item_state_event("Unittest_Humidity", 70)
        self.assertEqual("on_HighHumidity", self.humidity.state)
        tests.helper.oh_item.assert_value("Unittest_Output", "ON")

        # humidity below threshold again and timeout
        tests.helper.oh_item.item_state_event("Unittest_Humidity", 64)
        self.assertEqual("on_Extended", self.humidity.state)
        tests.helper.oh_item.assert_value("Unittest_Output", "ON")
        tests.helper.timer.call_timeout(self.transitions_timer_mock)
        self.assertEqual("off", self.humidity.state)
        tests.helper.oh_item.assert_value("Unittest_Output", "OFF")



import HABApp
import pydantic
import typing_extensions
from pydantic import model_validator

from habapp_rules.core.exceptions import HabAppRulesConfigurationError
from habapp_rules.core.pydantic_base import ConfigBase, ItemBase, ParameterBase


class EnergyMeterBaseItems(ItemBase):
    """Base class for energy meter items."""

    power_output: HABApp.openhab.items.NumberItem | None = pydantic.Field(None, description="power output item, unit is W")
    energy_output: HABApp.openhab.items.NumberItem | None = pydantic.Field(None, description="energy output item, unit is kWh")

    @pydantic.model_validator(mode="after")
    def validate_items(self) -> typing_extensions.Self:
        """Validate items.

        Returns:
            Validated items

        Raises:
            HabAppRulesConfigurationError: if items are not valid
        """
        if self.power_output is None and self.energy_output is None:
            msg = "At least one of power_output or energy_output must be set"
            raise HabAppRulesConfigurationError(msg)
        return self


class EnergyMeterSwitchItems(EnergyMeterBaseItems):
    """Items for virtual energy meter for a switch item."""

    monitored_switch: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="switch item, which will be monitored")


class EnergyMeterNumberItems(EnergyMeterBaseItems):
    """Items for virtual energy meter for a dimmer item."""

    monitored_item: HABApp.openhab.items.DimmerItem | HABApp.openhab.items.NumberItem = pydantic.Field(..., description="dimmer item, which will be monitored")


class EnergyMeterBaseParameter(ParameterBase):
    """Base class for energy meter parameters."""

    energy_update_resolution: int = pydantic.Field(0.010, description="update the energy item every x kWh. Default is 0.01kWh == 10 Wh", gt=0)


class EnergyMeterSwitchParameter(EnergyMeterBaseParameter):
    """Parameter for a virtual energy meter for a switch item."""

    power: float = pydantic.Field(..., description="typical power in W if switch is ON", gt=0)


class PowerMapping(pydantic.BaseModel):
    """Class to map a value to a power.

    This can be used to map e.g. a dimmer value to used power
    """

    value: float = pydantic.Field(..., description="dimmer / number value, which will be mapped to a power value")
    power: float = pydantic.Field(..., description="power in W")

    def __init__(self, value: float, power: float) -> None:
        """Init PowerMapping.

        Args:
            value: dimmer / number value
            power: power in W
        """
        super().__init__(value=value, power=power)


class EnergyMeterNumberParameter(EnergyMeterBaseParameter):
    """Parameter for a virtual energy meter for a dimmer item."""

    power_mapping: list[PowerMapping] = pydantic.Field(..., description="typical power if dimmed")

    @pydantic.field_validator("power_mapping")
    @classmethod
    def validate_power_mapping(cls, mappings: list[PowerMapping]) -> list[PowerMapping]:
        """Validate power_mapping.

        Args:
            mappings: list of power mappings

        Returns:
            Validated list of power mappings

        Raises:
            HabAppRulesConfigurationError: if power_mapping is not valid
        """
        if len(mappings) < 2:  # noqa: PLR2004
            msg = "power_mapping must have at least 2 elements"
            raise HabAppRulesConfigurationError(msg)

        mappings.sort(key=lambda x: x.value)
        return mappings

    def get_power(self, dimmer_value: float) -> float:
        """Get power depending on dimmer value.

        Args:
            dimmer_value: value of the dimmer

        Returns:
            power depending on dimmer value
        """
        idx = next((i for i, mapping in enumerate(self.power_mapping) if mapping.value >= dimmer_value), None)

        if idx is None:
            # If no such index is found, return the power of the last mapping
            return self.power_mapping[-1].power

        if idx == 0:
            # If the input value is less than the first mapping, return the power of the first mapping
            return self.power_mapping[0].power

        # Otherwise, interpolate the power between the two surrounding mappings
        prev_mapping = self.power_mapping[idx - 1]
        return prev_mapping.power + (self.power_mapping[idx].power - prev_mapping.power) * (dimmer_value - prev_mapping.value) / (self.power_mapping[idx].value - prev_mapping.value)


class EnergyMeterSwitchConfig(ConfigBase):
    """Config for virtual energy meter for a switch item."""

    items: EnergyMeterSwitchItems = pydantic.Field(..., description="items for the switch")
    parameter: EnergyMeterSwitchParameter = pydantic.Field(..., description="parameter for the switch")


class EnergyMeterNumberConfig(ConfigBase):
    """Config for virtual energy meter for a dimmer item."""

    items: EnergyMeterNumberItems = pydantic.Field(..., description="items for the dimmer")
    parameter: EnergyMeterNumberParameter = pydantic.Field(..., description="parameter for the dimmer")

    @model_validator(mode="after")
    def validate_model(self) -> typing_extensions.Self:
        """Validate model.

        Returns:
            Validated model

        Raises:
            HabAppRulesConfigurationError: if config is not valid
        """
        if isinstance(self.items.monitored_item, HABApp.openhab.items.DimmerItem):
            all_values = [mapping.value for mapping in self.parameter.power_mapping]
            if any(value < 0 for value in all_values) or any(value > 100 for value in all_values):  # noqa: PLR2004
                msg = "power_mapping values for dimmer items must be between 0 and 100"
                raise HabAppRulesConfigurationError(msg)
        return self



"""Rule for evaluating a humidity sensor."""

import logging
import typing

import HABApp

import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.sensors.config.humidity

LOGGER = logging.getLogger(__name__)


class HumiditySwitch(habapp_rules.core.state_machine_rule.StateMachineRule):
    """Rule for setting humidity switch if high humidity or a high humidity change is detected."""

    states: typing.ClassVar = [
        {"name": "off"},
        {
            "name": "on",
            "initial": "HighHumidity",
            "children": [
                {"name": "HighHumidity"},
                {"name": "Extended", "timeout": 99, "on_timeout": "on_extended_timeout"},
            ],
        },
    ]

    trans: typing.ClassVar = [
        {"trigger": "high_humidity_start", "source": "off", "dest": "on"},
        {"trigger": "high_humidity_start", "source": "on_Extended", "dest": "on_HighHumidity"},
        {"trigger": "high_humidity_end", "source": "on_HighHumidity", "dest": "on_Extended"},
        {"trigger": "on_extended_timeout", "source": "on_Extended", "dest": "off"},
    ]

    def __init__(self, config: habapp_rules.sensors.config.humidity.HumiditySwitchConfig) -> None:
        """Init humidity rule.

        Args:
            config: config for humidity switch rule
        """
        self._config = config
        habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, self._config.items.state)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self._config.items.humidity.name)

        # init state machine
        self._previous_state = None
        self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(model=self, states=self.states, transitions=self.trans, ignore_invalid_triggers=True, after_state_change="_update_openhab_state")
        self._set_initial_state()

        self.state_machine.get_state("on_Extended").timeout = self._config.parameter.extended_time

        # register callbacks
        self._config.items.humidity.listen_event(self._cb_humidity, HABApp.openhab.events.ItemStateUpdatedEventFilter())

    def _update_openhab_state(self) -> None:
        """Update OpenHAB state item. This should method should be set to "after_state_change" of the state machine."""
        super()._update_openhab_state()
        self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")

        self._set_output()
        self._previous_state = self.state

    def _set_output(self) -> None:
        """Set output."""
        target_state = "ON" if self.state in {"on_HighHumidity", "on_Extended"} else "OFF"
        habapp_rules.core.helper.send_if_different(self._config.items.output, target_state)

    def _get_initial_state(self, default_value: str = "initial") -> str:  # noqa: ARG002
        """Get initial state of state machine.

        Args:
            default_value: default / initial state

        Returns:
            if OpenHAB item has a state it will return it, otherwise return the given default value
        """
        return "on" if self._check_high_humidity() else "off"

    def _check_high_humidity(self, humidity_value: float | None = None) -> bool:
        """Check if humidity is above threshold.

        Args:
            humidity_value: humidity value, which should be checked. If None, the value of the humidity item will be used

        Returns:
             if humidity is above threshold
        """
        if humidity_value is None:
            if self._config.items.humidity.value is None:
                return False
            humidity_value = self._config.items.humidity.value

        return humidity_value >= self._config.parameter.absolute_threshold

    def _cb_humidity(self, event: HABApp.openhab.events.ItemStateUpdatedEvent) -> None:
        """Callback, which is triggered if the humidity was updated.

        Args:
            event: trigger event
        """
        if self._check_high_humidity(event.value):
            self.high_humidity_start()
        else:
            self.high_humidity_end()



"""Test config models for HCL light rules."""

import collections

import habapp_rules.actors.config.light_hcl
import habapp_rules.core.exceptions
import tests.helper.oh_item
import tests.helper.test_case_base


class TestLightHclConfig(tests.helper.test_case_base.TestCaseBase):
    """Test HCL config."""

    def test_sorted_color_config(self) -> None:
        """Test sorting of HCL values."""
        TestCase = collections.namedtuple("TestCase", "input, output")

        test_cases = [TestCase([(-1, 42), (0, 100), (1, 500)], [(-1, 42), (0, 100), (1, 500)]), TestCase([(0, 100), (-1, 42), (1, 500)], [(-1, 42), (0, 100), (1, 500)]), TestCase([(1, 500), (0, 100), (-1, 42)], [(-1, 42), (0, 100), (1, 500)])]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                parameter = habapp_rules.actors.config.light_hcl.HclElevationParameter(color_map=test_case.input)
                self.assertEqual(test_case.output, parameter.color_map)
